import Queue from "@ronomon/queue";
import request from "request-promise-native";
import cheerio from "cheerio";
import { map, forEach } from "async-array-methods";
import ImageDataURI from "image-data-uri";
import DB from "./db";

const DOMAIN = "https://www.ecb.europa.eu";

const CountryCollection = DB.getCollection("Country");
const CoinCollection = DB.getCollection("Coin");

let countryDataList = {};

const sizeToWorthInCentsConversion = {
    "1cent": 1,
    "2cents": 2,
    "5cents": 5,
    "10cents": 10,
    "20cents": 20,
    "50cents": 50,
    "1euro": 100,
    "2euro": 200
};

/**
 * Get country data by countryName
 * @param {string} countryName
 * @returns {Promise.<object>}
 */
async function getCountryData(countryName)
{
    if (countryDataList[countryName] === undefined) {
        const countryNameUrl = encodeURIComponent(countryName.replace("Vatican City", "Vatican"));
        const country = await request({
            uri: `https://restcountries.eu/rest/v2/name/${countryNameUrl}?fullText=true&fields=flag;region;subregion`,
            transform: function(body)
            {
                return JSON.parse(body)[0];
            },
        });
        countryDataList[countryName] = {
            FlagImageDataURI: await getDataUriFromImageUrl(country.flag),
            Region: country.region,
            SubRegion: country.subregion,
        };
    }
    return countryDataList[countryName];
}

/**
 * Get commemorative coin data by dataType
 * @param {object} dataWrapper
 * @param {string} dataType
 * @returns {string}
 */
function getCommemorativeCoinData(dataWrapper, dataType)
{
    return dataWrapper.children(".ecb-column").last().find("p > strong").filter(function() {
        return this.firstChild.data.includes(dataType);
    }).parent().contents().filter(function() {
        return this.type === "text";
    }).text().trim();
}

/**
 * Get country name from coin imageURL
 * @param {string} imageURL
 * @returns {string|undefined}
 */
function getCountryNameFromCoinImageUrl(imageURL)
{
    return Object.keys(countryDataList).find((countryName) => {
        let searchNames = [ countryName ];
        if (countryName === "Netherlands") {
            searchNames.push("Nederland");
        }
        return searchNames.some((name) => imageURL.includes(name));
    });
}

/**
 * Get dataURI for single image url
 * @param {string} url
 * @returns {Promise.<string>}
 */
async function getDataUriFromImageUrl(url)
{
    return ImageDataURI.encodeFromURL(url);
}

/**
 * Get Cheerio page object
 * @param {string} url
 * @returns {Promise.<*>}
 */
async function getCheerioObjectFromUrl(url)
{
    return await request({
        uri: url,
        transform: function(body)
        {
            return cheerio.load(body);
        },
    });
}

/**
 * Process standard coins page
 * @param {object} item
 * @param {function} $
 * @returns {Promise.<Array<object>>}
 */
async function processStandardCoinsPage(item, $)
{
    let coins = [];
    await forEach($(".coinDesc").get(), async (countryDataElement) => {
        const countryDataWrapper = $(countryDataElement);
        const imageUrls = countryDataWrapper.find("img.floatLeft").map((index, imageElement) => DOMAIN + imageElement.attribs.src).get();
        const countryName = countryDataWrapper.children("h3").text().trim();
        const description = countryDataWrapper.find("p").contents().filter(function() {
            return this.type === "text";
        }).text().trim();
        await map(imageUrls, async (imageUrl, index) => {
            coins.push({
                CountryName: countryName,
                Type: item.type,
                Size: item.size,
                WorthInCents: sizeToWorthInCentsConversion[item.size],
                CustomOrder: index,
                Description: description,
                Feature: null,
                Volume: null,
                Released: null,
                ImageURL: imageUrl,
                ImageDataURI: await getDataUriFromImageUrl(imageUrl),
            });
        }, { num: 8 });
    });
    return coins;
}

/**
 * Process commemorative coins page
 * @param {object} item
 * @param {function} $
 * @returns {Promise.<Array<object>>}
 */
async function processCommemorativeCoinsPage(item, $)
{
    let coins = [];
    const listWrapper = $(".coinDesc");
    await forEach(listWrapper.children("h2").get(), async (countryNameElement, index) => {
        let countryName = $(countryNameElement).text().trim();
        const dataWrapper = listWrapper.children(".ecb-columns.ecb-twoCol.clearFix").eq(index);
        if (countryName !== "Euro area countries") {
            if (countryName === "Vatican") {
                countryName = "Vatican City";
            }
            if (countryName === "The Netherlands") {
                countryName = "Netherlands";
            }
            let url;
            coins.push({
                CountryName: countryName,
                Type: item.type,
                Size: item.size,
                WorthInCents: sizeToWorthInCentsConversion[item.size],
                CustomOrder: null,
                Description: getCommemorativeCoinData(dataWrapper, "Description"),
                Feature: getCommemorativeCoinData(dataWrapper, "Feature"),
                Volume: getCommemorativeCoinData(dataWrapper, "Issuing volume"),
                Released: new Date(getCommemorativeCoinData(dataWrapper, "Issuing date")).getTime(),
                ImageURL: url = DOMAIN + dataWrapper.find("img.floatLeft").attr("src"),
                ImageDataURI: await getDataUriFromImageUrl(url),
            });
        } else {
            await forEach(dataWrapper.find("img.floatLeft").get(), async (imageElement) => {
                const imageURL = DOMAIN + imageElement.attribs.src;
                coins.push({
                    CountryName: getCountryNameFromCoinImageUrl(imageURL),
                    Type: item.type,
                    Size: item.size,
                    WorthInCents: sizeToWorthInCentsConversion[item.size],
                    CustomOrder: null,
                    Description: getCommemorativeCoinData(dataWrapper, "Description"),
                    Feature: getCommemorativeCoinData(dataWrapper, "Feature"),
                    Volume: getCommemorativeCoinData(dataWrapper, "Issuing volume"),
                    Released: new Date(getCommemorativeCoinData(dataWrapper, "Issuing date")).getTime(),
                    ImageURL: imageURL,
                    ImageDataURI: await getDataUriFromImageUrl(imageURL),
                });
            });
        }
    });
    return coins;
}

/**
 * Save coin data to database
 * @param {object} coinData
 * @param {string} coinData.CountryName
 * @param {string} coinData.CountryID Used internally
 * @param {string} coinData.Type
 * @param {string} coinData.Size
 * @param {number} coinData.WorthInCents
 * @param {number} coinData.CustomOrder
 * @param {string} coinData.Description
 * @param {string} coinData.Feature
 * @param {string} coinData.Volume
 * @param {string} coinData.Released
 * @param {string} coinData.ImageURL
 * @param {string} coinData.ImageDataURI
 * @returns {Promise.<void>}
 */
async function saveCoinDataToCollection(coinData)
{
    let countryData = await getCountryData(coinData.CountryName);
    countryData.Name = coinData.CountryName;
    const result = await CountryCollection.update(
        { Name: coinData.CountryName },
        { $set: countryData },
        { upsert: true }
    );
    delete coinData.CountryName;
    coinData.CountryID = result.updated[0]._id;
    await CoinCollection.update(
        {
            CountryID: coinData.CountryID,
            ImageURL: coinData.ImageURL,
        },
        { $set: coinData },
        { upsert: true }
    );
}

export default class DataRefresh
{
    /**
     * @param {function} onChangeStats
     */
    constructor(onChangeStats)
    {
        this._totalItems = 0;
        this._doneItems = 0;
        this._onChangeStatsCallback = onChangeStats;
        this._queue = new Queue(2);
        this._isDone = new Promise((resolve, reject) => {
            this._queue.onEnd = (error) => {
                if (error) {
                    reject(error.toString());
                }
                resolve();
            };
        });
        this._queue.onData = async (item, done) => {
            await this._processItem(item);
            done();
        };
    }

    /**
     * Process queue item
     * @param {object} item
     * @param {string} item.type
     * @param {string} item.size
     * @param {string} item.url
     * @returns {Promise.<void>}
     * @private
     */
    async _processItem(item)
    {
        const $ = await getCheerioObjectFromUrl(item.url);
        let coins;
        if (item.type === "standard") {
            coins = await processStandardCoinsPage(item, $);
        } else if ("commemorative") {
            coins = await processCommemorativeCoinsPage(item, $);
        } else {
            throw new Error("Unknown coin type!");
        }
        await forEach(coins, async (coinData) => {
            await saveCoinDataToCollection(coinData);
        });
        this._doneItems++;
        this._notifyStats();
    }

    /**
     * Add item to queue
     * @param {object} item
     * @param {string} item.type
     * @param {string} item.size
     * @param {string} item.url
     * @private
     */
    _addItem(item)
    {
        this._queue.push(item);
        this._totalItems++;
        this._notifyStats();
    }

    /**
     * Notify stats callback about current total and done items
     * @private
     */
    _notifyStats()
    {
        this._onChangeStatsCallback(this._totalItems, this._doneItems);
    }

    /**
     * Refresh data
     * @returns {Promise.<void>}
     */
    async refresh()
    {
        const sizes = [ "1cent", "2cents", "5cents", "10cents", "20cents", "50cents", "1euro", "2euro" ];
        sizes.forEach((size) => {
            this._addItem({
                type: "standard",
                size: size,
                url: DOMAIN + `/euro/coins/${size}/html/index.en.html`,
            });
        });
        for (let year = 2004; year <= new Date().getUTCFullYear(); year++) {
            this._addItem({
                type: "commemorative",
                size: "2euro",
                url: DOMAIN + `/euro/coins/comm/html/comm_${year}.en.html`,
            });
        }
        this._queue.end();
        await this._isDone;
    }
}